package me.msoucy.ptures.controller

import me.msoucy.ptures.model.Creature
import me.msoucy.ptures.model.KnockedOut
import me.msoucy.ptures.model.Skill
import me.msoucy.ptures.model.Target
import me.msoucy.ptures.model.Team
import me.msoucy.ptures.view.BattleView
import me.msoucy.ptures.view.SkillChoice

sealed class BattleType(vararg val teams: Team) {
    init {
        assert(teams.isNotEmpty())
    }

    abstract val numGroups: Int
}

sealed class IndividualBattle(vararg teams: Team) : BattleType(*teams) {
    override val numGroups: Int
        get() = teams.size
}

class SingleBattle(vararg teams: Team) : IndividualBattle(*teams)
class DoubleBattle(vararg teams: Team) : IndividualBattle(*teams)
class TripleBattle(vararg teams: Team) : IndividualBattle(*teams)
class TeamBattle(override val numGroups: Int, vararg teams: Team) : BattleType(*teams)
class RaidBattle(val boss: Creature, vararg teams: Team) : BattleType(*teams) {
    override val numGroups: Int
        get() = 2
}


class Engine(private val battle: BattleType) {

    private var currentCreature = 0

    private val creatures get() = battle.teams.flatMap { it.activeCreatures }

    val attacker: Creature
        get() {
            return creatures[currentCreature].creature
        }

    private val forcedSkills = mutableMapOf<Creature, Skill>()

    private val activeCreatures
        get() = creatures
                .filter { !it.creature.hasStatus<KnockedOut>() }
                .sortedBy { it.creature.spd }

    fun resolveTurn(view: BattleView) {
        forcedSkills.clear()
        // All preconditions
        for (i in activeCreatures.indices) {
            val creature = creatures[i]
            for (status in creature.creature.statuses) {
                status.onTurnStart(this, creature.creature)
            }
        }
        // Get the moves each creature will use this turn
        val moves = activeCreatures.indices.map { i ->
            val forcedSkill = forcedSkills[creatures[i].creature]
            if (forcedSkill != null) {
                SkillChoice(forcedSkill, creatures[nextOpponent(i)].creature)
            } else {
                creatures[i].chooseSkill(activeCreatures.map { it.creature })
            }
        }
        // Resolve each move
        for (i in activeCreatures.indices) {
            // Resolve move
            val (skill, target) = moves[i]
            currentCreature = i
            for (step in skill.damageSteps) {
                for (targetCreature in getTargetList(step.target, target)) {
                    if (attacker.hits(targetCreature, step)) {
                        targetCreature.apply(step, attacker)
                        step.applyStatus(targetCreature)
                    }
                }
            }
            for (step in skill.postSteps) {
                for (targetCreature in getTargetList(step.target, target)) {
                    step.apply(targetCreature)
                }
            }
        }
        // All post conditions
        for (creatureView in activeCreatures) {
            val creature = creatureView.creature
            for (status in creature.statuses) {
                status.onTurnEnd(this)
            }
        }
    }

    private fun getTargetList(target: Target, selected: Creature): List<Creature> {
        return when (target) {
            Target.Self -> listOf(attacker)
            Target.Selected -> listOf(selected)
            Target.Others -> creatures.filter { it.creature != attacker }.map { it.creature }
            Target.Opponents -> creatures.filter { it.playerId != creatures[currentCreature].playerId }.map { it.creature }
            Target.All -> creatures.map { it.creature }
        }
    }

    private fun nextOpponent(idx: Int): Int {
        var nextIdx = idx + 1
        while (nextIdx != idx) {
            if (creatures[nextIdx].playerId != creatures[idx].playerId) {
                return nextIdx
            }
            nextIdx = (nextIdx + 1) % creatures.size
        }
        // There are no opponents... so use the "none" index
        return -1
    }
}